/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file canny.cpp
@ingroup ExamplePrograms
@brief Example program for QVision library. Does XXXXXXXXXXXXXXX

¡ Canny !

\code
	./canny --URL=<path or url to video>
\endcode

\todo write about canny worker.

@param URL path to video file, or video source identificator. See @ref QVMPlayerCamera.
@author PARP Research Group. University of Murcia, Spain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <QDebug>

#include <qvcore/qvapplication.h>
#include <qvmplayercamera/qvmplayercamera.h>
#include <qvgui/qvisioninterface.h>

#include <qvdta/qvpolyline.h>
#include <qvdta/qvcontour.h>

/// \cond
class CannyWorker: public QVWorker
	{
	public:
		CannyWorker(QString name): QVWorker(name)
			{
			addProperty<double>("Threshold high", inputFlag,	150,	"High threshold for Canny operator", 50, 1000);
			addProperty<double>("Threshold low", inputFlag,		50,	"Low threshold for Canny operator", 10, 500);
			addProperty< QVImage<uChar,1> >("Canny image", outputFlag);
			addProperty< QVImage<uChar,1> >("Input image", inputFlag|outputFlag);
			addProperty< QList<qvdta::QVPolyline> >("Contour list", outputFlag);
			}

		void iterate()
			{
			//////////////////
			// Read input parameters
			QVImage<uChar,1> image = getPropertyValue< QVImage<uChar,1> >("Input image");
			uInt cols = image.getCols(), rows = image.getRows();
			QVImage<sFloat>	imageFloat(cols, rows), dX(cols, rows), dY(cols, rows), dXNeg(cols, rows);
			QVImage<uChar>	canny(cols, rows), buffer;
		
			//////////////////
			// Convert image from uChar to sShort
			qvipp::Convert(image, imageFloat);
			timeFlag("Convert image from uChar to sShort");
		
			//////////////////
			// Obtain horizontal and vertical gradients from image
			qvipp::FilterSobelHorizMask(imageFloat,dY,3);
			qvipp::FilterSobelVertMask(imageFloat,dX,3);
			qvipp::MulC(dX, dXNeg, -1);
			timeFlag("Obtain horizontal and vertical gradients from image");
		
			/////////////////
			// Apply Canny operator
			qvipp::CannyGetSize(canny, buffer);
			qvipp::Canny(dXNeg, dY, canny, buffer, getPropertyValue<double>("Threshold low"), getPropertyValue<double>("Threshold high"));
			timeFlag("Apply Canny operator");

			/////////////////
			// Get contours
			QList< qvdta::QVPolyline> contourList = qvdta::getLineContoursThreshold4Connectivity(canny, 128);
			timeFlag("Get contours");
		
			/////////////////
			// Publish resulting data
			setPropertyValue< QVImage<uChar,1> >("Canny image",canny);
			setPropertyValue< QList< qvdta::QVPolyline> >("Contour list", contourList);
			timeFlag("Publish resulting images");
			}
	};

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		
		"Example program for QVision library. Gets component tree from images."

		);
	
	CannyWorker cannyWorker("Canny");

	QVMPlayerCamera camera("Video");
	camera.link(&cannyWorker,"Input image");

	QVisionInterface interface;

	QVImageCanvas imageCanvas;
	imageCanvas.linkProperty(cannyWorker,"Canny image");
	imageCanvas.linkProperty(cannyWorker,"Contour list");

	return app.exec();
	}

/// \endcond
	
