/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvpropertyholder.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVPROPERTYHOLDER_H
#define QVPROPERTYHOLDER_H

#include <QStringList>
#include <QVariant>
#include <QRegExp>
#include <QSet>
#include <QReadWriteLock>
#include <QSemaphore>
#include <QDebug>

#include <iostream>

#include <qvcore/qvapplication.h>

#define READ_PROPERTY(HOLDER, TYPE, NAME, VARIABLE) TYPE VARIABLE = HOLDER->getPropertyValue<TYPE>(NAME);
#define READ_PROPERTY_MAX(HOLDER, TYPE, NAME, VARIABLE) TYPE VARIABLE = HOLDER->getPropertyMaximum<TYPE>(NAME);
#define READ_PROPERTY_MIN(HOLDER, TYPE, NAME, VARIABLE) TYPE VARIABLE = HOLDER->getPropertyMinimum<TYPE>(NAME);

/// @class QVPropertyHolder qvpropertyholder.h qvcore/qvpropertyholder.h
/// @brief Base class for dynamic property container objects.
///
/// Parent class for objects to which we want to add execution-time accessible
/// properties. Holds methods to register new properties, read and set their
/// values, and to get human readable information about them (useful to show
/// a related help text). It also supports minimum and maximum checking for 
/// integer and double values, parsing from the command line for QString, bool,
/// integer and double types.
///
/// This introspection functionality is in a sense similar to that given by
/// QObject, but it does not inherit from it, so we can have properties for 
/// classes which need not contain the Q_OBJECT macro, thus not needing to be
/// forcibly in a .h file. It also allows for direct access of the QVApplication
/// to variables created by the final programmer, in order to, for example, show
/// associated widgets to property so that the user can directly access to them.
///
/// Also, one of the most important features that this introspection adds to the
/// QVision framework is the ability to link properties among different property
/// holders, in the sense that the "input properties" of one of them can be
/// "linked" to be read in execution time from the "output properties" of
/// another holder. The main application of this feature is to be able to
/// program independently different threads of execution (QVThread), which can
/// be later used and "linked" afterwards by other programmers, thus making the
/// programmed QVThread easily reusable.
///
/// \todo Parámetro --help de línea de comandos no muestra propiedades de los workers.
///
/// @ingroup qvcore
class QVPropertyHolder
	{
	public:
		/// @brief Semantics for a link between QVPropertyHolders.
		///
		/// If a link between a source property of a QVPropertyHolder object A
		/// and a corresponding destination property of another one B is
		/// declared as synchronous, it means that every time that the
		/// readInputProperties() method is executed on B, the system assures
		/// that one (and just one) corresponding call to the
		/// writeOutputProperties() has been previously executed on A.
		/// In the (common) case of a link between two QVThread's, that would
		/// mean, for example, that each running of the source QVThread
		/// corresponds exactly to another (posterior) running of the
		/// destination QVThread.
		///
		/// That is equivalent to say that every data produced by the source is
		/// always consumed by the destination, so no data is ever "lost" in
		/// that sense. The obvious cost is that both threads will have to run
		/// at the speed of the slowest, of course.
		///
		/// The alternative is an asynchronous link, in which the only thing
		/// that the system assures on the properties read by the destination
		/// is that, in the case of a multiple link (several input properties
		/// of B connected to several corresponding output properties of B), all
		/// of them correspond to the same iteration of the source (more
		/// precisely, to the last execution of the writeOutputProperties()
		/// method in A, independently of the specific sequence number of that
		/// iteration). We say that the output is 'coherent' in all the
		/// properties linked this way. But, as has been already said,
		/// the iterations are not linked in the temporal sense, so, for
		/// example, for the case of QVThreads, both could run asynchronously,
		/// and at different speeds.
		///
		/// See also the QVPropertyHolder::linkProperty() method.
		typedef enum {AsynchronousLink,SynchronousLink} LinkType;

		/// @brief Flags for each property.
		///
		/// Currently, if a property is of the "input type" and/or "output type"
		/// are the only supported flags.
		typedef enum {noInputOutputFlag=0x0,inputFlag=0x1,outputFlag=0x2} PropertyInputOutputFlag;
		Q_DECLARE_FLAGS(PropertyInputOutputFlags, PropertyInputOutputFlag)

		/// \cond
		typedef enum {noLinkFlag=0x0,linkedInputFlag=0x1,linkedOutputFlag=0x2} PropertyLinkFlag;
		Q_DECLARE_FLAGS(PropertyLinkFlags, PropertyLinkFlag)
		/// \endcond

		/// @brief Constructs a QVPropertyHolder with a given string identifier.
		///
		/// @param name Identifier for the object to be created.
		QVPropertyHolder(const QString name = QString());

		/// @brief Virtual destructor.
		///
		/// Frees the QVPropertyHolder, and all its associated links. Besides,
		/// this destructor can (and in many cases should!) also be
		/// reimplemented in subclasses.
		virtual ~QVPropertyHolder();

		/// @brief Sets a new string identifier for a QVPropertyHolder.
		///
		/// @param name New string identifier for the object.
		void setName(const QString name);

		/// @brief Retrieves identifier QString of the object.
		/// @return QString identifying the property holder object.
		const QString getName() const;

		/// @brief Gets list of properties of the QVPropertyHolder.
		///
		/// Retrieves the names of the properties registered in the property
		/// holder, in a QStringList.
		/// @return QStringList of names of properties in the object.
		QList<QString> getPropertyList() const;

		/// @brief Gets the names of the properties of a given type.
		///
		/// Retrieves the names of those properties, registered in the property
		/// holder, which hold values of the type given by the template
		/// parameter, in a QStringList.
		/// @return QList of identifying QStrings.
		template <class Type> QList<QString> getPropertyListByType() const
			{
			QList<QString> result;
			QList<QString> names = variants.keys();

			for(QList<QString>::iterator i = names.begin();i != names.end();++i)
				if(isType<Type>(*i))
					result.append(*i);

			return result;
			}

		/// @brief Checks if a property is of a given type.
		///
		/// Returns true if type passed as template parameter equals type
		/// of value contained in a property.
		/// @param name QString identifying the property.
		/// @param ok Pointer to bool value which is set to true if property
		///           exists, and to false otherwise.
		/// @return True if property has same type as parameter template.
		template <class Type> bool isType(QString name,bool *ok = NULL) const
			{
			if(not checkExists(name,"QVPropertyHolder::propertyIsType()"))
				{
				if(ok != NULL) *ok = FALSE;
				return FALSE;
				}
			if(ok != NULL) *ok = TRUE;
			QVariant::Type type = QVariant::fromValue(Type()).type();
			if ((type != QVariant::UserType) && (variants[name].type() == type))
				return TRUE;
			if (variants[name].userType() == QVariant::fromValue(Type()).userType())
				return TRUE;
			return FALSE;
			}

		/// @brief Check whether the property holder has a given property.
		/// @param name QString identifying the property.
		/// @return True if the QVPropertyHolder contains a property with the
		///         given name, otherwise returns false.
		bool containsProperty(const QString name) const;

		/// @brief Given a property identifying QString (see getPropertyList() 
		///        function above), returns the type of the property.
		/// @param name QString identifying the property.
		/// @param ok Pointer to bool value which is set to true if property
		///           exists, and to false otherwise.
		/// @return a <A HREF="http://doc.trolltech.com/qtopia4.2/qvariant.html#Type-enum">QVariant::Type</A>
		/// value. Note that user defined properties will return
		/// QVariant::UserType value.
		QVariant::Type getPropertyType(const QString name, bool *ok = NULL) const;

		/// @brief Registers a new property in the property holder.
		///
		/// @param name Name of the new property.
		/// @param value Initial value for the property.
		/// @param info Info string for the property.
		/// @param flags Whether the property is input, output, or both.
		/// @return True if property was correctly added, false otherwise.
		template <class Type> bool addProperty(const QString name,
			const PropertyInputOutputFlags flags = inputFlag,
			const Type & value = Type(), const QString info = QString("(Info not available)"))
			{
			/*if(not checkIsNewProperty(name,"QVPropertyHolder::addProperty()"))
				return FALSE;
			insertion_order.push_back(name);
			_info[name] = info;
			io_flags[name] = flags;
			link_flags[name] = noLinkFlag;

			// If possible, we set the initial value from the command line;
			// otherwise, we let the default:
			variants[name] = QVariant::fromValue(value);
			setPropertyFromArguments<Type>(name);
			return TRUE;*/
			if (addPropertyFromQVariant(name, flags, QVariant::fromValue(value), info))
				setPropertyFromArguments<Type>(name);
			else
				return false;

			return true;
			}

		/// @brief Registers a new property in the property holder.
		///
		/// @param name Name of the new property.
		/// @param value Initial value for the property.
		/// @param info Info string for the property.
		/// @param flags Whether the property is input, output, or both.
		/// @param minValue Minimum value for the property.
		/// @param maxValue Maximum value for the property.
		/// @return True if property was correctly added, false otherwise.
		template <class Type> bool addProperty(const QString name,
			const PropertyInputOutputFlags flags,
			const Type & value, const QString info,
			const Type & minValue, const Type & maxValue)
			{
			addProperty<Type>(name, flags, value, info);
			setPropertyRange2<Type>(name, minValue, maxValue);
			return TRUE;
			}

		/// @brief De-registers a property in the property holder.
		///
		/// @param name Name of the property to delete.
		/// @return True if property was correctly removed, false otherwise.
		bool removeProperty(const QString name);

		/// \deprecated Should set range in the constructor.
		bool setPropertyRange(const QString name, const double & minimum, const double & maximum);

		/// \deprecated Should set range in the constructor.
		bool setPropertyRange(QString name, int & minimum, int & maximum);

		/// @brief Tells if a range has been set for the property.
		///
		/// @param name Name of the property.
		/// @return True if there exist a range for the property.
		bool hasRange(const QString name);

		/// @brief Tells if a property is of the input type.
		///
		/// @param name Name of the property.
		/// @return True if the property is of the input type.
		bool isInput(const QString name);

		/// @brief Returns flags for a property
		///
		/// @param name Name of the property.
		/// @return Flags for the property.
		PropertyInputOutputFlags getPropertyFlags(const QString name) { return io_flags[name]; }


		/// @brief Tells if a property is of the output type.
		///
		/// @param name Name of the property.
		/// @return True if the property is of the output type.
		bool isOutput(const QString name);

		/// @brief Tells if a property is of the input type and is linked.
		///
		/// @param name Name of the property.
		/// @return True if the property is of the input type and is linked.
		bool isLinkedInput(const QString name);

		/// @brief Tells if a property is of the output type and is linked.
		///
		/// @param name Name of the property.
		/// @return True if the property is of the output type and is linked.
		bool isLinkedOutput(const QString name);

		/// @brief Sets value for property.
		///
		/// If the property is not writable, this function will return false,
		/// and won't change property value. Besides, if there is a
		/// [minimum,maximum] range and the value is not in it, also fails to
		/// set the value. In both cases the function issues a warning through
		/// stderr, and returns false.
		/// @param name Identifying QString for the property.
		/// @param value Value to set the property with.
		/// @return True if the value was set successfully.
		template <class Type> bool setPropertyValue(const QString name, const Type &value)
			{
			if(not checkExists(name,"QVPropertyHolder::setPropertyValue()"))
				return FALSE;
			else if (not correctRange(name,value))
				return FALSE;
			else	{
				variants[name] = QVariant::fromValue<Type>(value);
				return TRUE;
				}
			}

		/// @brief Method to get the actual value of a property.
		///
		/// @param name Identifying QString for the property.
		/// @param ok Pointer to bool value which is set to true if property
		///           exists, and to false otherwise.
		/// @return Actual value for the property.
		template <class Type> Type getPropertyValue(const QString name, bool *ok = NULL) const
			{
			if (not checkExists(name,"QVPropertyHolder::getPropertyValue()"))
				if(ok != NULL) *ok = FALSE;
			else
				if(ok != NULL) *ok = TRUE;
			return variants[name].value<Type>();
			}

		/// @brief Method to get the value of a property as a QVariant.
		///
		/// @param name Identifying QString for the property.
		/// @param ok Pointer to bool value which is set to true if property
		///           exists, and to false otherwise.
		/// @return Actual value for the property.
		QVariant getPropertyQVariantValue(const QString name, bool *ok = NULL) const;

		/// @brief Method to get the maximum value for a property.
		///
		/// @param name Identifying QString for the property.
		/// @param ok Pointer to bool value which is set to true if property
		///           exists, and to false otherwise.
		/// @return Maximum value for the property.
		template <class Type> Type getPropertyMaximum(const QString name, bool *ok = NULL) const
			{
			if(not checkExists(name,"QVPropertyHolder::getPropertyMaximum()"))
				if(ok != NULL) *ok = FALSE;
			else if(not maximum.contains(name) and not minimum.contains(name))
				{
				QString str = QString("QVPropertyHolder::getPropertyMaximum():")
							  + QString(" property ") + name
							  + QString(" has no maximum value in ")
							  + QString("holder ") + getName() + QString(".");
				setLastError(str);
				if(qvApp->isRunning()) {
					std::cerr << qPrintable("Warning: " + str + "\n");
				} // Otherwise, qApp will show the error and won't start the program.
				if(ok != NULL) *ok = FALSE;
				}
			else
				if(ok != NULL) *ok = TRUE;
			return maximum[name].value<Type>();
			}

		/// @brief Method to get the minimum value for a property.
		///
		/// @param name Identifying QString for the property.
		/// @param ok Pointer to bool value which is set to true if property
		///           exists, and to false otherwise.
		/// @return Minimum value for the property.
		template <class Type> Type getPropertyMinimum(const QString name, bool *ok = NULL) const
			{
			if(not checkExists(name,"QVPropertyHolder::getPropertyMinimum()"))
				if(ok != NULL) *ok = FALSE;
			else if(not maximum.contains(name) and not minimum.contains(name))
				{
				QString str = QString("QVPropertyHolder::getPropertyMinimum():")
							  + QString(" property ") + name
							  + QString(" has no minimum value in ")
							  + QString("holder ") + getName() + QString(".");
				setLastError(str);
				if(qvApp->isRunning()) {
					std::cerr << qPrintable("Warning: " + str + "\n");
				} // Otherwise, qApp will show the error and won't start the program.
				if(ok != NULL) *ok = FALSE;
				}
			else
				if(ok != NULL) *ok = TRUE;
			return minimum[name].value<Type>();
			}

		/// @brief Method to get the info QString for int typed property.
		///
		/// @param name Identifying QString for the property.
		/// @param ok Pointer to bool value which is set to true if property
		///           exists, and to false otherwise.
		/// @return The info QString about the property, if defined. Empty
		///         QString otherwise.
		QString getPropertyInfo(const QString name, bool *ok = NULL) const;

		/// @brief Gets the (possible) last error string.
		///
		/// @return A QString indicating the last error that occurred in an
		///         operation on the the property, if defined. Empty QString
		///         otherwise.
		QString getLastError() const;

		/// @brief Formatted description of properties (useful to show help).
		///
		/// Obtains a human readable description QString of all the object's
		/// properties that have some associated info (otherwise they are
		/// ignored).
		/// @return QString containing human readable list of properties, and
		///         info about them.
		const QString infoInputProperties() const;

		/// @brief Links properties among QVPropertyHolder's.
		///
		/// This method safely links properties among property holders in a
		/// thread safe manner. It is mainly used when programming
		/// multithreaded applications -with multiple QVWorkers-, but also, for
		/// example, when linking workers (running in an independent thread)
		/// with GUI widgets (running in the main thread).
		///
		/// The pointer this corresponds to the source QVPropertyHolder.
		///
		/// See also QVPropertyHolder::LinkType.
		/// @param sourcePropertyName Identifying QString for the source property.
		/// @param destinyHolder Pointer to the destination QVPropertyHolder.
		/// @param destinyPropertyName Identifying QString for the destination property.
		/// @param linkType Synchronous or asynchronous.
		/// @return TRUE if the connection went OK, FALSE otherwise (for example,
		/// due to type mismatch, some of the properties does not exists, or
		/// whatever).
		bool linkProperty(QString sourcePropertyName, QVPropertyHolder *destinyHolder, QString destinyPropertyName, LinkType linkType);

		/// @brief Completely unlinks a property holder.
		///
		/// Deletes all the input and output links associated to the holder.
		/// Of course, all the corresponding necessary unlinks in other property
		/// holders are also performed.
		void unlink();

	protected:

		template <class Type> bool setPropertyRange2(const QString name, const Type & minimum, const Type & maximum)
			{
			if(not checkExists(name,"QVPropertyHolder::setPropertyRange()"))
				return FALSE;
			if(minimum <= getPropertyValue<Type>(name) and
				maximum >= getPropertyValue<Type>(name))
				{
					this->minimum[name] = QVariant::fromValue(minimum);
					this->maximum[name] = QVariant::fromValue(maximum);
					return TRUE;
				} else {
					QString str =  "QVPropertyHolder::setPropertyRange(): property " +
								   name + " in holder " + getName() + " has value " +
								   QString("%1").arg(getPropertyValue<Type>(name)) +
								   ", which is not valid for the range [" +
								   QString("%1,%2").arg(minimum).arg(maximum) + "]." ;
					setLastError(str);
					if(qvApp->isRunning()) {
						std::cerr << qPrintable("Warning: " + str + "\n");
					} // Otherwise, qApp will show the error and won't start the program.
					return FALSE;
				}
			}

		/// @brief Read linked input properties from other QVPropertyHolder's.
		///
		/// This is a protected function that should be called from the
		/// subclasses of QVPropertyHolder just before starting to use any
		/// input property linked to another property holder (for
		/// example, just before the worker() function in a QVWorker, or just
		/// before the viewer() function in a QVCanvas).
		void readInputProperties();

		/// @brief Write linked output properties to other QVPropertyHolder's.
		///
		/// This is a protected function that should be called from the
		/// subclasses of QVPropertyHolder just after having a new set of
		/// output properties ready to be read by another property holder (for
		/// example, just after the worker() function in a QVWorker, or just
		/// after the viewer() function in a QVCanvas). The effect is that
		/// a coherent state of the linked output properties is saved, so
		/// other QVPropertyHolder's can safely read from it -while we can
		/// be writing new values corresponding to a new iteration-.
		void writeOutputProperties();

		/// @brief Parses a property value from a string.
		///
		/// Parses a QString containing the value for a property (mainly useful
		/// when initializing it from the command line). QVPropertyHolder
		/// implements a series of default parsers (namely for bool, int, double
		/// and QString properties). But the method is templatized to allow
		/// customization of the parsed QString, if we wish to add new parameter
		/// types in the subclasses. In that case, users should specialize
		/// the corresponding template this way:\n
		/// \code
		/// template <> bool MyDerivedQVPropertyHolder::parseArgument<MyType>(const QString parameter, const QString value)
		/// {
		///     bool ok;
		///     MyType myVar1 = myParserForMyType1(value,&ok);
		///     if(ok) { // parsing OK 
		///          setPropertyValue<MyType>(parameter,myVar1);
		///          return TRUE;
		///     } else { // parsing failed
		///          QString errorString = "MyDerivedQVPropertyHolder::parseArgument(): holder "
		///          + getName() + ": value " + value
		///          + " is not a valid MyType value for parameter " + parameter + ".\n";
		///          setLastError(errorString);
		///          return FALSE;
		///     }
		/// }
		/// \endcode
		/// The format of the input params in the command line should be the
		/// following: \n
		///	\c --[holder:]\<name\>=\<value\>
		/// \n where
		///	- \b holder is the (optional) name of the property holder.
		///	- \b name is the name of the property.
		///	- \b value is the value to initialize the property.
		/// @param parameter The name of the property to be set.
		/// @param value QString from which to obtain the value.
		///
		/// If the optional holder name is not present, any holder which has
		/// a property with name 'name' is assigned the corresponding parsed
		/// value.
		template <class Type> bool parseArgument(const QString parameter, const QString value);

		/// @brief Sets an error string message.
		///
		/// Sets an error string message corresponding to the last error that
		/// occurred in an operation on the QVPropertyHolder (if such an error
		/// occurred). If no error occurred since the creation of the object,
		/// the string will be empty. The method is const because it only
		/// modifies the error message (which is mutable), not the object
		/// itself.
		void setLastError(QString str) const;

		/// @brief Registers a new property in the property holder.
		///
		/// @param name Name of the new property.
		/// @param flags Whether the property is input, output, or both.
		/// @param variant Initial QVariant value for the property
		/// @param info Info string for the property.
		/// @return True if property was correctly added, false otherwise.
		bool addPropertyFromQVariant(const QString &name, const PropertyInputOutputFlags flags, QVariant variant, const QString info)
			{
			if(not checkIsNewProperty(name,"QVPropertyHolder::addProperty()"))
				return FALSE;
			insertion_order.push_back(name);
			_info[name] = info;
			io_flags[name] = flags;
			link_flags[name] = noLinkFlag;

			variants[name] = variant;
			return TRUE;
			}


	private:
		QString name;
		mutable QString errorString;
		QMap<QString, QVariant> variants,safelyCopiedVariants;
		QMap<QString, QVariant> minimum, maximum;
		QMap<QString, QString> _info;
		QMap<QString, PropertyInputOutputFlags> io_flags;
		QMap<QString, PropertyLinkFlags> link_flags;
		QList<QString> insertion_order;


		QReadWriteLock RWLock;
		class QVPropertyHolderLink {
		  public:
			QVPropertyHolderLink(QVPropertyHolder *_qvp_orig,QString _prop_orig,QVPropertyHolder *_qvp_dest,QString _prop_dest,LinkType _link_type) : qvp_orig(_qvp_orig), prop_orig(_prop_orig), qvp_dest(_qvp_dest), prop_dest(_prop_dest), link_type(_link_type), markedForDeletion(FALSE) {
				// SyncSemaphoreIn value must initially be 1:
				SyncSemaphoreIn.release();
				// SyncSemaphoreOut is simply initialized with 0 (default value).
			};
			QVPropertyHolder *qvp_orig;
			QString prop_orig;
			QVPropertyHolder *qvp_dest;
			QString prop_dest;
			LinkType link_type;
			QSemaphore SyncSemaphoreIn,SyncSemaphoreOut;
			bool markedForDeletion;
		};
		QMap<QString, QVPropertyHolderLink* > inputLinks;
		QMap<QString, QList<QVPropertyHolderLink*> > outputLinks;

		template <class Type> bool setPropertyFromArguments(QString propertyName)
			{
			QStringList arguments = qvApp->arguments();

			QMutableStringListIterator iterator(arguments);
			while (iterator.hasNext())
				{
				QString argument = iterator.next();
				// Only if the name of the property holder is the same as defined in
				// the argument, in the form --name:property=value, (with name:
				// optional) the value will be intended to be applied to the property:
				if (argument.contains(QRegExp("^--")))
					{
					QString propertyHolderName(argument), parameter(argument), value(argument);
					// Checks whether in the argument string there is a name for
					// restricting the application of the value to a variable of
					// a specific property holder.
					if (argument.contains(QRegExp("^--[^=]+:")))
						{
						propertyHolderName.remove(QRegExp("^--")).remove(QRegExp(":.*"));
						if(propertyHolderName != getName())
							continue;
						parameter.remove(QRegExp("^--[^=]+:")).remove(QRegExp("=.*$"));
						}
					else
						{
						parameter.remove(QRegExp("^--")).remove(QRegExp("=.*$"));
						}
					if(parameter != propertyName)
						continue;
					// If we get here, then we must assign the value to the property:
					value.remove(QRegExp("^--[^=]*="));
					if(parseArgument<Type>(parameter,value))
						{
						if(not isInput(propertyName))
							{
							QString str = QString("QVPropertyHolder::setPropertyFromArguments():")
										+ QString(" property ") + propertyName
							  			+ QString(" in holder ") + getName()
										+ QString(" is not of Input type, and cannot be parsed.");
							setLastError(str);
							//qApp will show the error and won't start the program.
							}
						qvApp->setArgumentAsUsed(argument);
						return TRUE;
						}
					}
				}
				// If we get here, then we did not parse any value for the property:
				return FALSE;
			}

		bool correctRange(const QString name, const double & value) const;
		// Any type inference ambiguity resolved with this:
		// 1. For int values:
		bool correctRange(const char *name, const int & value) const;
		bool correctRange(QString name, const int & value) const;

		// 2. Rest of types have no range, so we always return true:
		template <typename T> bool correctRange(const QString parameter, const T & value) {
			Q_UNUSED(parameter);
			Q_UNUSED(value);
			return TRUE;
		}

		bool checkExists(const QString name, const QString methodname) const;
		bool checkIsNewProperty(const QString name, const QString methodname) const;
	};

template <class Type> bool QVPropertyHolder::parseArgument(const QString parameter, const QString value)
	{
		errorString = "QVPropertyHolder::parseArgument(): holder " + getName() +
						": parameter " + parameter +
						" has an unknown type to command line parser " +
						QString("(trying to parse value %1)").arg(value) ;
		return FALSE;
	}

template <> bool QVPropertyHolder::parseArgument<bool>(const QString parameter, const QString value);

template <> bool QVPropertyHolder::parseArgument<int>(const QString parameter, const QString value);

template <> bool QVPropertyHolder::parseArgument<double>(const QString parameter, const QString value);

template <> bool QVPropertyHolder::parseArgument<QString>(const QString parameter, const QString value);

Q_DECLARE_OPERATORS_FOR_FLAGS(QVPropertyHolder::PropertyInputOutputFlags)
Q_DECLARE_OPERATORS_FOR_FLAGS(QVPropertyHolder::PropertyLinkFlags)

#endif
